#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <pwd.h>

#include <libxml/tree.h>
#include <libxml/parser.h>

#include <glib.h>

#include <utillib.h>

#include "downmand-marshal.h"

#include "main.h"
#include "download.h"
#include "project.h"


GType           project_get_type            ();
static void     project_class_init          (ProjectClass *klass);
static void     project_init                (Project *project);

void        project_download_exit       (Download *download, gpointer data);
gboolean    project_download_real_exit  (GIOChannel *source,
                                         GIOCondition condition,
                                         gpointer data);

void    project_cleanup                     (void *arg);
void    project_session_batch_to_xml        (xmlNodePtr xmlnode, List *batch,
                                             gchar *name);
void    project_session_download_to_xml     (xmlNodePtr xmlnode,
                                             Download *download);

gint    project_session_load                (Project *project);
void    project_session_xml_to_batch        (Project *project,
                                             xmlNodePtr xmlnode);
void    project_session_xml_to_download     (Project *project, List *batch,
                                             xmlNodePtr xmlnode);

Download *  project_download_new        (Project *project, gchar *url);
void        project_download_update     (Download *download, gpointer data);

void    project_download_links          (Project *project, Download *download);


enum {
    PROJECT_PATH,
    PROJECT_AUTODOWNLOAD,
    PROJECT_MAXDOWNLOADS,
    DOWNLOAD_NEW,
    DOWNLOAD_STOP,
    DOWNLOAD_DONE,
    DOWNLOAD_DELETE,
    LAST_SIGNAL
};

static gint project_signals[LAST_SIGNAL] = { 0 };

static gpointer parent_class = NULL;


GType project_get_type ()
{
    static GType project_type = 0;

    if (!project_type)
    {
        static const GTypeInfo project_info =
        {
            sizeof (ProjectClass),
            (GBaseInitFunc) NULL,
            (GBaseFinalizeFunc) NULL,
            (GClassInitFunc) project_class_init,
            NULL,           /* class_finalize */
            NULL,           /* class_data */
            sizeof (Project),
            0,              /* n_preallocs */
            (GInstanceInitFunc) project_init
        };

        project_type = g_type_register_static (G_TYPE_OBJECT,
                                               "Project",
                                               &project_info,
                                               0);
    }

    return project_type;
}

static void project_class_init (ProjectClass *klass)
{
    GObjectClass *object_class;

    object_class = (GObjectClass*) klass;

    parent_class = g_type_class_peek_parent (klass);

    project_signals[PROJECT_AUTODOWNLOAD] = g_signal_new (
            "project_autodownload_changed",
            G_OBJECT_CLASS_TYPE (klass),
            G_SIGNAL_RUN_FIRST,
            G_STRUCT_OFFSET (ProjectClass, project_autodownload_changed),
            NULL, NULL,
            downmand_marshal_VOID__VOID,
            G_TYPE_NONE,
            0);

    project_signals[PROJECT_MAXDOWNLOADS] = g_signal_new (
            "project_maxdownloads_changed",
            G_OBJECT_CLASS_TYPE (klass),
            G_SIGNAL_RUN_FIRST,
            G_STRUCT_OFFSET (ProjectClass, project_maxdownloads_changed),
            NULL, NULL,
            downmand_marshal_VOID__VOID,
            G_TYPE_NONE,
            0);

    project_signals[DOWNLOAD_NEW] = g_signal_new ("download_new",
            G_OBJECT_CLASS_TYPE (klass),
            G_SIGNAL_RUN_FIRST,
            G_STRUCT_OFFSET (ProjectClass, download_new),
            NULL, NULL,
            downmand_marshal_VOID__OBJECT,
            G_TYPE_NONE,
            1, download_get_type());

    project_signals[DOWNLOAD_STOP] = g_signal_new ("download_stop",
            G_OBJECT_CLASS_TYPE (klass),
            G_SIGNAL_RUN_FIRST,
            G_STRUCT_OFFSET (ProjectClass, download_stop),
            NULL, NULL,
            downmand_marshal_VOID__OBJECT,
            G_TYPE_NONE,
            1, download_get_type());

    project_signals[DOWNLOAD_DONE] = g_signal_new ("download_done",
            G_OBJECT_CLASS_TYPE (klass),
            G_SIGNAL_RUN_FIRST,
            G_STRUCT_OFFSET (ProjectClass, download_done),
            NULL, NULL,
            downmand_marshal_VOID__OBJECT,
            G_TYPE_NONE,
            1, download_get_type());

    project_signals[DOWNLOAD_DELETE] = g_signal_new ("download_delete",
            G_OBJECT_CLASS_TYPE (klass),
            G_SIGNAL_RUN_FIRST,
            G_STRUCT_OFFSET (ProjectClass, download_delete),
            NULL, NULL,
            downmand_marshal_VOID__OBJECT,
            G_TYPE_NONE,
            1, download_get_type());

    klass->project_autodownload_changed = NULL;
    klass->project_maxdownloads_changed = NULL;
    klass->download_new = NULL;
    klass->download_stop = NULL;
    klass->download_done = NULL;
    klass->download_delete = NULL;
}

static void
project_init (Project *project)
{
    project->hash = g_hash_table_new (NULL, NULL);
    project->hash_active = g_hash_table_new (NULL, NULL);
	project->batch = list_new();
	project->batch_active = list_new();
    project->download_id = 1;

    // Counters
    project->queued_downloads = 0;

    /* init mutexes */
    g_static_mutex_init (&project->batch_mutex);
    g_static_mutex_init (&project->batch_active_mutex);
}

Project *
project_new (guint project_id, GError **error)
{
    Project *project = PROJECT (g_object_new (project_get_type (), NULL));
    gint listener[2];

    if (pipe (listener) < 0)
    {
//FIXME: Better error handling
        return NULL;
    }

    project->listener_read = g_io_channel_unix_new (listener[0]);
    project->listener_write = g_io_channel_unix_new (listener[1]);
    g_io_channel_set_encoding (project->listener_read, NULL, NULL);
    g_io_channel_set_encoding (project->listener_write, NULL, NULL);

    project->listen_tag = g_io_add_watch (project->listener_read,
                    G_IO_IN | G_IO_PRI | G_IO_ERR | G_IO_HUP | G_IO_NVAL,
                    project_download_real_exit, project);

    project->project_id = project_id;

    /* default configuration */
    project->max_downloads = 2;
    project->auto_download = 1;

    return project;
}

void
project_scheduler (Project *project)
{
    GHashTable *hash_active = project->hash_active;
	List *batch = project->batch;
	List *batch_active = project->batch_active;
	Node *element;
    GError *err = NULL;

    project_lock_batch (project);
    project_lock_active (project);

	// prepare for the posible iteration
	element = list_begin(batch);

	while ((project->auto_download)
		&& (list_size(batch_active) < project->max_downloads)
		&& (project->queued_downloads > 0))
	{
		Download *download;

		// search for a queued download
		while((element != 0))
		{
			download = list_node_data(element);
			if (download_status_get(download) == STATUS_QUEUED)
				break;
			else
				element = list_next(element);
		}

		if (element != 0)
		{
			Node *new_element;

			// check per server downloads limit
			// start a new download.

			new_element = list_node_new(download);
			list_add(batch_active, new_element);
            g_hash_table_insert (hash_active,
                                 GINT_TO_POINTER (download_id_get (download)),
                                 new_element);
            download_queue_set (download, DOWNLOAD_QUEUE_ACTIVE);
            download_status_set (download, STATUS_INIT);
			project->queued_downloads--;

			// Start a new thread to download it
			download_start(download, &err);

            if (err != NULL)
            {
                download_queue_set (download, DOWNLOAD_QUEUE_STOPPED);
                download_status_set (download, STATUS_STOPPED);
                list_remove (batch_active, new_element);
                g_hash_table_remove (hash_active,
                                     GINT_TO_POINTER (download_id_get (download)));

g_print ("%s\n", err->message);
                g_error_free (err);
                err = NULL;
                g_signal_emit (G_OBJECT (project),
                               project_signals[DOWNLOAD_STOP],
                               0,
                               download);
            }

			// and move to the next element
			element = list_next(element);
		}
	}

    /* Do speed scheduling */
    if ((project->speedlimit) && (list_size (batch_active)))
    {
        GList *downloads_list = NULL;
        GList *glist;
		Download *download;
        Download *slowest_download;
        guint active_downloads;
        guint bandwidth_left;
        guint bandwidth_piece;

        active_downloads = list_size (batch_active);
        bandwidth_left = project->speedlimit;
        bandwidth_piece = project->speedlimit / active_downloads;

        /* copy the active downloads list */
        for (element = list_begin (batch_active); element != 0;
             element = list_next (element))
        {
            download = list_node_data(element);
            downloads_list = g_list_prepend (downloads_list, download);
        }

        /* Remove slowest_download and keep recalculating bandwidth_piece */
        do {
            slowest_download = NULL;
            for (glist = g_list_first (downloads_list); glist != 0;
                 glist = g_list_next (glist))
            {
                download = (Download *) glist->data;
                if (download_speedlimit_get (download))
                {
                    if (slowest_download == NULL)
                        slowest_download = download;
                    else if (download_speedlimit_get (download)
                             < download_speedlimit_get (slowest_download))
                        slowest_download = download;
                }
            }

            if (slowest_download != NULL)
            {
                if (download_speedlimit_get (slowest_download)
                    <= bandwidth_piece)
                {
                    bandwidth_left -= download_speedlimit_get (
                                                            slowest_download);
                    downloads_list = g_list_remove (downloads_list,
                                                    slowest_download);
                    download_real_speedlimit_set (slowest_download,
                                    download_speedlimit_get (slowest_download));
                    /* recompute bandwidth_piece */
                    active_downloads--;
                    if (active_downloads)
                        bandwidth_piece = bandwidth_left / active_downloads;
                }
                else
                {
                    /* No more iterations needed, get out */
                    slowest_download = NULL;
                }
            }
        } while (slowest_download && active_downloads);

        /* look for already limited downloads */
        for (glist = g_list_first (downloads_list); glist != 0;
             glist = g_list_next (glist))
        {
            download = (Download *) glist->data;
            download_real_speedlimit_set (download, bandwidth_piece);
        }
    }
    else
    {
		Download *download;

        /* Set real_speedlimit on limited downloads and clear the rest */
        for (element = list_begin (batch_active); element != 0;
             element = list_next (element))
        {
            download = list_node_data(element);
            if (download_speedlimit_get (download))
            {
                download_real_speedlimit_set (download,
                                            download_speedlimit_get (download));
            }
            else
                download_real_speedlimit_set (download, 0);
        }
    }

    project_unlock_active (project);
    project_unlock_batch (project);
}

guint
project_id (Project *project)
{
    return project->project_id;
}

const char *project_name(Project *project)
{
	return project->name;
}

guint
project_download_id (Project *project)
{
    return project->download_id++;
}

gint
project_command_download_new (Project *project, gchar *command,
                              GError **error)
{
    Download *download;
    Node *element;
    gchar **commands;
    gint i = 0;
    guint command_id = 0, value = 0;

    download = project_download_new (project, NULL);

//FIXME: use GError here
	if (download == 0)
		return 1;

    download_log_add (download, "* Download created.\n");

    download_queue_set (download, DOWNLOAD_QUEUE_BATCH);
    download_status_set (download, STATUS_QUEUED);
    download_path_set (download, project->default_path);
    download_default_path_set (download, TRUE);

    commands = g_strsplit (command, "|", 0);

    for (command = commands[i]; command != NULL; command = commands[++i])
    {
        if (sscanf(command, "%u ", &command_id) == 0)
        {
            download_log_add (download, "* Bad command id: %s\n", command);
            continue;
        }

        switch (command_id)
        {
            /* Download commands */
            case DM_COMMAND_DOWNLOAD_URL:
//FIXME: invalid utf
//                download_log_add (download, "* Set url: %s\n", command + 4);
                download_url_set (download, command + 4);
                break;

            case DM_COMMAND_DOWNLOAD_PATH:
                download_log_add (download, "* Set PATH: %s\n", command + 4);
                download_path_set (download, command + 4);
                download_default_path_set (download, FALSE);
                break;

            case DM_COMMAND_DOWNLOAD_STOP:
                if (sscanf (command + 4, "%u", &value) > 0)
                {
                    if (value)
                    {
                        download_log_add (download, "* Set status: stopped.\n");
                        download_queue_set (download, DOWNLOAD_QUEUE_STOPPED);
                        download_status_set (download, STATUS_STOPPED);
                    }
                }
                break;

            case DM_COMMAND_DOWNLOAD_RECURSIVE:
                if (sscanf (command + 4, "%u", &value) > 0)
                {
                    download_log_add (download,
                                  "* Set FLAGS: Recursive on: %u.\n", value);
                    download_recursive_set (download, value);
                }
                break;

            case DM_COMMAND_DOWNLOAD_SPEED:
                if (sscanf (command + 4, "%u", &value) > 0)
                {
                    download_log_add (download,
                                  "* Set FLAGS: Speed limit on: %u.\n", value);
                    download_speedlimit_set (download, value);
                }
                break;

            default:
                download_log_add (download, "COMMAND NOT UNDERSTOOD: %s\n", command);
                break;
        }
    }

    g_strfreev(commands);

	element = list_node_new(download);

    project_lock_batch (project);

    download_log_add (download, "* Added to project list.\n");
	list_add(project->batch, element);
    g_hash_table_insert (project->hash,
                         GINT_TO_POINTER (download_id_get (download)),
                         element);
    if (download_status_get (download) == STATUS_QUEUED)
        project->queued_downloads++;

    project_unlock_batch (project);

    g_signal_emit (G_OBJECT (project), project_signals[DOWNLOAD_NEW], 0,
                   download);

	project_scheduler(project);

	return 0;
}


/* This function is called when a download exits.
 */
void
project_download_exit (Download *download, gpointer data)
{
    Project *project = PROJECT (data);
    static GStaticMutex mutex = G_STATIC_MUTEX_INIT;
    gsize bytes_written;

    download_log_add (download, "* download_exit called: thread = %p\n",
                      g_thread_self ());
    g_static_mutex_lock (&mutex);
    g_io_channel_write_chars (project->listener_write, (gchar *) &download,
                              sizeof (Download *),
                              &bytes_written,
                              NULL);
    g_io_channel_flush (project->listener_write, NULL);
    g_static_mutex_unlock (&mutex);
}

gboolean
project_download_real_exit (GIOChannel *source, GIOCondition condition,
                            gpointer data)
{
    Project *project = PROJECT (data);
    Download *download;
	GHashTable *hash = project->hash;
	GHashTable *hash_active = project->hash_active;
	List *batch = project->batch;
	List *batch_active = project->batch_active;
	Node *element;
    GError *err = NULL;
    gsize bytes_read;

    g_io_channel_read_chars (project->listener_read, (gchar *) &download,
                             sizeof (Download *),
                             &bytes_read,
                             NULL);

    if (!IS_DOWNLOAD (download))
        return TRUE;

    download_log_add (download, "* download_real_exit called: thread = %p\n",
                      g_thread_self ());

    element = g_hash_table_lookup (hash_active,
                               GINT_TO_POINTER (download_id_get (download)));
    if (element)
	{
        download_log_add (download, "* Download exitted.\n");
		switch (download_exit_cause(download))
		{
			case DOWNLOAD_EXIT_STOPPED:
                download_log_add (download, "* Stop. Move from active to stopped.\n");
                project_lock_active (project);
				list_remove(batch_active, element);
                g_hash_table_remove (hash_active,
                                 GINT_TO_POINTER (download_id_get (download)));
                download_queue_set (download, DOWNLOAD_QUEUE_STOPPED);
                project_unlock_active (project);
                g_signal_emit (G_OBJECT (project),
                               project_signals[DOWNLOAD_STOP],
                               0,
                               download);
				break;

			case DOWNLOAD_EXIT_QUEUED:
                download_log_add (download, "* Queue. Move from active to batch.\n");
                project_lock_active (project);
				list_remove(batch_active, element);
                g_hash_table_remove (hash_active,
                                 GINT_TO_POINTER (download_id_get (download)));
                download_queue_set (download, DOWNLOAD_QUEUE_BATCH);
                project_unlock_active (project);
                g_signal_emit (G_OBJECT (project),
                               project_signals[DOWNLOAD_STOP],
                               0,
                               download);
				break;

			case DOWNLOAD_EXIT_DELETE:
                download_log_add (download, "* Delete. Remove from active.\n");
                project_lock_active (project);
				list_remove(batch_active, element);
                g_hash_table_remove (hash_active,
                                 GINT_TO_POINTER (download_id_get (download)));
				list_node_delete(element);
                project_unlock_active (project);

                element = g_hash_table_lookup (hash,
                               GINT_TO_POINTER (download_id_get (download)));
                if (element)
				{
                    download_log_add (download, "* Remove from project.\n");
                    project_lock_batch (project);
					list_remove(batch, element);
				    g_hash_table_remove (hash,
                                GINT_TO_POINTER (download_id_get (download)));
					list_node_delete(element);
                    project_unlock_batch (project);
				}

                g_signal_emit (G_OBJECT (project),
                               project_signals[DOWNLOAD_DELETE],
                               0,
                               download);
				g_object_unref (G_OBJECT (download));
				break;

			case DOWNLOAD_EXIT_DONE:
                download_log_add (download, "* Done. Move to done.\n");
                project_lock_active (project);
				list_remove(batch_active, element);
                g_hash_table_remove (hash_active,
                                GINT_TO_POINTER (download_id_get (download)));
                download_queue_set (download, DOWNLOAD_QUEUE_DONE);
                project_unlock_active (project);
                g_signal_emit (G_OBJECT (project),
                               project_signals[DOWNLOAD_DONE],
                               0,
                               download);

                if (download->recursive)
                    project_download_links (project, download);
				break;

            case DOWNLOAD_EXIT_SHUTDOWN:
                download_log_add (download, "* Downmand shutdown.\n");
                break;

			default:
                download_log_add (download, "* No reason to exit. Trying again.\n");
//FIXME: check for number of retries
				download_start(download, &err);

                if (err != NULL)
                {
                    project_lock_active (project);
                    download_queue_set (download, DOWNLOAD_QUEUE_STOPPED);
                    download_status_set (download, STATUS_STOPPED);
                    list_remove (batch_active, element);
                    g_hash_table_remove (hash_active,
                                GINT_TO_POINTER (download_id_get (download)));

                    g_error_free (err);
                    project_unlock_active (project);
                    g_signal_emit (G_OBJECT (project),
                                   project_signals[DOWNLOAD_STOP],
                                   0,
                                   download);
                }

				break;
		}
	}

    return TRUE;
}

gint
project_command_download_start (Project *project, guint id, GError **error)
{
	GHashTable *hash = project->hash;
	GHashTable *hash_active = project->hash_active;
	List *batch_active = project->batch_active;
	Node *element;
    GError *err = NULL;

    project_lock_batch (project);
    project_lock_active (project);

    element = g_hash_table_lookup (hash, GINT_TO_POINTER (id));
    if (element)
	{
		Download *download = list_node_data(element);
		Node *new_element;

        if (download->url)
        {
            switch(download_queue_get (download))
            {
                case DOWNLOAD_QUEUE_BATCH:
                    project->queued_downloads--;
                case DOWNLOAD_QUEUE_STOPPED:
                    new_element = list_node_new(download);
                    list_add(batch_active, new_element);
                    g_hash_table_insert (hash_active,
                                GINT_TO_POINTER (download_id_get (download)),
                                new_element);
                    download_queue_set (download, DOWNLOAD_QUEUE_ACTIVE);
                    download_start(download, &err);

                    if (err != NULL)
                    {
                        download_queue_set (download, DOWNLOAD_QUEUE_STOPPED);
                        download_status_set (download, STATUS_STOPPED);
                        list_remove (batch_active, new_element);
                        g_hash_table_remove (hash_active,
                                GINT_TO_POINTER (download_id_get (download)));

g_print ("%s\n", err->message);
                        g_error_free (err);
                    }
                    break;

			default:
				break;
            }
        }
	}

    project_unlock_active (project);
    project_unlock_batch (project);

	return 0;
}

gint
project_command_download_stop (Project *project, guint id, GError **error)
{
	GHashTable *hash = project->hash;
	Node *element;

    project_lock_batch (project);

    element = g_hash_table_lookup (hash, GINT_TO_POINTER (id));

    if (element)
	{
		Download *download = list_node_data(element);

        switch(download_queue_get (download))
		{
			case DOWNLOAD_QUEUE_BATCH:
                download_log_add (download, "* Stop. Move from batch to stopped.\n");
				if (download_status_get(download) == STATUS_QUEUED)
					project->queued_downloads--;
                download_status_set (download, STATUS_STOPPED);
				download_queue_set (download, DOWNLOAD_QUEUE_STOPPED);
                g_signal_emit (G_OBJECT (project),
                               project_signals[DOWNLOAD_STOP],
                               0,
                               download);
				break;

			case DOWNLOAD_QUEUE_ACTIVE:
				download_log_add (download, "* Stopping the download.\n");
                download_stop(download, DOWNLOAD_STOP_NONE);
				break;

			default:
                download_log_add (download, "* Stop: can't be stopped.\n");
				break;
		}
	}

    project_unlock_batch (project);

	return 0;
}

gint
project_command_download_queue (Project *project, guint id, GError **error)
{
	GHashTable *hash = project->hash;
	Node *element;

    project_lock_batch (project);

    element = g_hash_table_lookup (hash, GINT_TO_POINTER (id));

	if (element)
	{
		Download *download = list_node_data(element);

		switch(download_queue_get (download))
		{
			case DOWNLOAD_QUEUE_STOPPED:
                download_log_add (download, "* Queue. Move from stopped to batch.\n");
				project->queued_downloads++;
                download_status_set (download, STATUS_QUEUED);
                download_queue_set (download, DOWNLOAD_QUEUE_BATCH);
                g_signal_emit (G_OBJECT (project),
                               project_signals[DOWNLOAD_STOP],
                               0,
                               download);
				break;

			case DOWNLOAD_QUEUE_ACTIVE:
                download_log_add (download, "* Queuing the download.\n");
				project->queued_downloads++;
				download_stop(download, DOWNLOAD_STOP_QUEUE);
				break;

			default:
                download_log_add (download, "* Queue: can't be queued.\n");
				break;
		}
	}

    project_unlock_batch (project);

	return 0;
}

gint
project_command_download_delete(Project *project, guint id, GError **error)
{
	GHashTable *hash = project->hash;
	List *batch = project->batch;
	Node *element;

    project_lock_batch (project);

    element = g_hash_table_lookup (hash, GINT_TO_POINTER (id));

	if (element)
	{
		Download *download = list_node_data(element);

		switch(download_queue_get (download))
		{
			case DOWNLOAD_QUEUE_BATCH:
				if (download_status_get (download) == STATUS_QUEUED)
					project->queued_downloads--;
			case DOWNLOAD_QUEUE_STOPPED:
                download_log_add (download, "* Delete. Removing from project.\n");
				list_remove(batch, element);
			    g_hash_table_remove (hash, GINT_TO_POINTER (id));
				list_node_delete(element);

                g_signal_emit (G_OBJECT (project),
                               project_signals[DOWNLOAD_DELETE],
                               0,
                               download);

				g_object_unref (G_OBJECT (download));
				break;

			case DOWNLOAD_QUEUE_ACTIVE:
                download_log_add (download, "* Deleting.\n");
				download_stop(download, DOWNLOAD_STOP_DELETE);
				break;

			case DOWNLOAD_QUEUE_DONE:
                download_log_add (download, "* Delete. Removing from project.\n");
				list_remove(batch, element);
			    g_hash_table_remove (hash, GINT_TO_POINTER (id));
                g_signal_emit (G_OBJECT (project),
                               project_signals[DOWNLOAD_DELETE],
                               0,
                               download);
				g_object_unref (G_OBJECT (download));
				list_node_delete(element);
				break;

			default:
                download_log_add (download, "* Delete: invalid queue.\n");
				break;
		}
	}
    else
    {
//FIXME: log to project linelog
        g_print("Download id %i not found, so not deleted!\n", id);
    }

    project_unlock_batch (project);

	return 0;
}

void
download_link (gpointer key, gpointer value, gpointer data)
{
    Project *project = PROJECT (data);
    Download *download;
    Node *element;
    gchar *url = (gchar *) key;
    guint recursive;

    recursive = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (project),
                                                    "recursive"));

g_print("%s\n", url);
    download = project_download_new (project, url);
    download_queue_set (download, DOWNLOAD_QUEUE_BATCH);
    download_status_set (download, STATUS_QUEUED);
    download_path_set (download, project->default_path);
    download_default_path_set (download, TRUE);
    download_recursive_set (download, recursive);
	element = list_node_new(download);

    project_lock_batch (project);

	list_add(project->batch, element);
    g_hash_table_insert (project->hash,
                         GINT_TO_POINTER (download_id_get (download)),
                         element);
    if (download_status_get (download) == STATUS_QUEUED)
        project->queued_downloads++;

    project_unlock_batch (project);

    g_signal_emit (G_OBJECT (project), project_signals[DOWNLOAD_NEW], 0,
                   download);

//FIXME: really?
//	project_scheduler(project);
}

void
project_download_links (Project *project, Download *download)
{
    GHashTable *hash;
//FIXME    GHashTable *hash_filtered;
    GString *buffer = g_string_new (NULL);
    gchar *base;
    gchar *last;
    guint recursive = download->recursive;

    if (recursive-- == 0)
        return;

    if ((!download->url->path) ||
        (download->url->path[strlen (download->url->path) - 1] == '/'))
    {
        g_string_printf(buffer, "%s/%s",
                        download->local_path, "index.html");
    }
    else
    {
        g_string_printf(buffer, "%s/%s",
                        download->local_path,
                        strrchr (download->url->path,'/'));
    }

    base = g_strdup (download->url_string);
    last = strrchr (base, '/');
    last[1] = '\0';

    hash = parse_html_file (buffer->str, base);

//FIXME: do this for all patterns
//    move_matching_strings (hash, hash_filtered, pattern);

    g_object_set_data (G_OBJECT (project), "recursive",
                       GINT_TO_POINTER (recursive));
    g_hash_table_foreach (hash, download_link, project);
}

int project_session_save(Project *project)
{
	List *batch = project->batch;
	xmlDocPtr doc;
	GString *buffer = g_string_new(NULL);

	doc = xmlNewDoc("1.0");
	doc->children = xmlNewDocNode(doc, NULL, "DownloadManager", NULL);

    project_lock_batch (project);

	project_session_batch_to_xml(doc->children, batch, "batch");

    project_unlock_batch (project);

//FIXME: save to a different file and test for integrity
	g_string_printf(buffer, "%s/batch", project->path);

	if (xmlSaveFile(buffer->str, doc) == -1)
	{
// FIXME: better error handling
		return 1;
	}

	xmlFreeDoc(doc);
	g_string_free(buffer, TRUE);

	return 0;
}

void project_session_batch_to_xml(xmlNodePtr xmlnode, List *batch, char *name)
{
	Download *download;
	Node *element;
	xmlNodePtr newxml;

	newxml = xmlNewChild(xmlnode, NULL, name, NULL);

	element = list_begin(batch);
	while (element != 0)
	{
		download = list_node_data(element);

		project_session_download_to_xml(newxml, download);

		element = list_next(element);
	}
}

void project_session_download_to_xml(xmlNodePtr xmlnode, Download *download)
{
	xmlNodePtr newxml;
	GString *buffer = g_string_new(NULL);

	newxml = xmlNewChild(xmlnode, NULL, "Download", NULL);

    g_string_printf(buffer, "%u", download_queue_get (download));
    xmlSetProp(newxml, "queue", buffer->str);

    g_string_printf(buffer, "%u", download_status_get (download));
    xmlSetProp(newxml, "status", buffer->str);

    g_string_printf(buffer, "%u", download_recursive_get (download));
    xmlSetProp(newxml, "recursive", buffer->str);

	xmlSetProp(newxml, "url", download_url_string(download));

    g_string_printf(buffer, "%lu", download_bytes_total_get (download));
	xmlSetProp(newxml, "bytes_total", buffer->str);

    g_string_printf(buffer, "%lu", download_bytes_done_get (download));
    xmlSetProp(newxml, "bytes_done", buffer->str);

	g_string_free(buffer, TRUE);
}

int project_session_load(Project *project)
{
	xmlDocPtr doc;
	xmlNodePtr node;
	GString *buffer = g_string_new(NULL);

	g_string_printf(buffer, "%s/batch", project->path);
	doc = xmlParseFile(buffer->str);

	if (!doc)
		return 1;

	if (!doc->children || !doc->children->name
		|| strcasecmp(doc->children->name, "DownloadManager") != 0)
	{
		xmlFreeDoc(doc);
		return 1;
	}

	for (node = doc->children->children; node != NULL; node = node->next)
	{
		project_session_xml_to_batch(project, node);
	}

	xmlFreeDoc(doc);
	g_string_free(buffer, TRUE);

	return 0;
}

void project_session_xml_to_batch(Project *project, xmlNodePtr xmlnode)
{
	xmlNodePtr node;
	List *batch;

	if (!xmlnode->name)
		return;

	if (strcasecmp(xmlnode->name, "batch") == 0)
	{
		batch = project->batch;
	}
	else
	{
		g_warning("Unknow queue\n");
		return;
	}

	for (node = xmlnode->children; node != NULL; node = node->next)
	{
		project_session_xml_to_download(project, batch, node);
	}
}

void project_session_xml_to_download(Project *project, List *batch,
	xmlNodePtr xmlnode)
{
	Download *download;
	Node *element;
    Node *new_element;
	char *url;
	DownloadQueue queue;
	enum dm_download_status status;
	unsigned long int bytes_total, bytes_done;
    guint recursive = 0;
    xmlChar *prop;
    GError *err = NULL;

	if (!xmlnode->name || strcasecmp(xmlnode->name, "Download") != 0)
		return;

	url = xmlGetProp(xmlnode, "url");
	if (url == 0)
		return;

	queue = atol(xmlGetProp(xmlnode, "queue"));
	status = atol(xmlGetProp(xmlnode, "status"));

	recursive = atol(xmlGetProp(xmlnode, "status"));

	download = project_download_new (project, url);
	if (download == 0)
		return;

    prop = xmlGetProp(xmlnode, "bytes_total");
    if (prop != NULL)
    {
        bytes_total = atol(prop);
        download_bytes_total_set (download, bytes_total);
        xmlFree(prop);
    }

//FIXME: if it is a normal download, check that it equals the size of the file
    prop = xmlGetProp(xmlnode, "bytes_done");
    if (prop != NULL)
    {
        bytes_done = atol(prop);
        download_bytes_done_set (download, bytes_done);
        xmlFree(prop);
    }

	element = list_node_new(download);

    project_lock_batch (project);

    download_queue_set (download, queue);
    download_status_set (download, status);
    download_path_set (download, project->default_path);
    download_default_path_set (download, TRUE);
    download_recursive_set (download, recursive);
	list_add(batch, element);
    g_hash_table_insert (project_hash(project),
                         GINT_TO_POINTER (download_id_get (download)),
                         element);

    switch (queue)
    {
        case DOWNLOAD_QUEUE_ACTIVE:
            new_element = list_node_new(download);
            list_add (project->batch_active, new_element);
            g_hash_table_insert (project->hash_active,
                                 GINT_TO_POINTER (download_id_get (download)),
                                 new_element);
            download_start (download, &err);

            if (err != NULL)
            {
                download_queue_set (download, DOWNLOAD_QUEUE_STOPPED);
                download_status_set (download, STATUS_STOPPED);
                list_remove (project->batch_active, new_element);
                g_hash_table_remove (project->hash_active,
                                GINT_TO_POINTER (download_id_get (download)));

                g_error_free (err);
            }
            break;

        case DOWNLOAD_QUEUE_BATCH:
            download_status_set (download, STATUS_QUEUED);
            project->queued_downloads++;
            break;

        case DOWNLOAD_QUEUE_STOPPED:
            download_status_set (download, STATUS_STOPPED);
            break;

        default:
            break;
    }

    project_unlock_batch (project);
}

Download *project_download_new(Project *project, gchar *url)
{
    Download *download = download_new (url);

    if (download != NULL)
    {
        g_signal_connect (download, "download_update",
                          G_CALLBACK (project_download_update),
                          project);

        g_signal_connect (download, "download_exit",
                          G_CALLBACK (project_download_exit),
                          project);
    }

    download_id_set (download, project_download_id(project));
    download_logdir_set (download, project->path);

    return download;
}

/* This function is called when a download changes its configuration.
 */
void project_download_update(Download *download, gpointer data)
{
/*    Project *project = PROJECT (data);*/
	GString *buffer = g_string_new(NULL);

    g_print("Got signal download_update from download %i\n", download->id);

    g_string_free(buffer, TRUE);
}

int project_preferences_save(Project *project)
{
	xmlDocPtr doc;
	xmlNodePtr newxml;
	GString *buffer = g_string_new(NULL);

//FIXME: modified flag --> do not write the config file if there are no modifications
	doc = xmlNewDoc("1.0");
	doc->children = xmlNewDocNode(doc, NULL, "DownloadManagerPrefs", NULL);

	newxml = xmlNewChild(doc->children, NULL, "Settings", NULL);
    if (project->default_path)
    {
        xmlSetProp(newxml, "DefaultDir", project->default_path);
    }
    else
    {
        g_string_printf(buffer, "%s/downloads/%s/", g_get_home_dir(),
                        project->name);
        xmlSetProp(newxml, "DefaultDir", buffer->str);
    }

	g_string_printf(buffer, "%i", project->max_downloads);
	xmlSetProp(newxml, "MaxDownloads", buffer->str);

	g_string_printf(buffer, "%i", project->auto_download);
	xmlSetProp(newxml, "AutoDownload", buffer->str);

//FIXME: save to a different file and test for integrity
	g_string_printf(buffer, "%s/prefs", project->path);

	if (xmlSaveFile(buffer->str, doc) == -1)
	{
// FIXME: better error handling
		return 1;
	}
	xmlFreeDoc(doc);

	return 0;
}

int project_preferences_load(Project *project)
{
	xmlDocPtr doc;
	xmlNodePtr node;
	GString *buffer = g_string_new(NULL);
	xmlChar *prefs;

	g_string_printf(buffer, "%s/prefs", project->path);
	doc = xmlParseFile(buffer->str);

	if (!doc)
		return 1;

	if (!doc->children || !doc->children->name
		|| strcasecmp(doc->children->name, "DownloadManagerPrefs") != 0)
	{
		xmlFreeDoc(doc);
		return 1;
	}

	node = doc->children->children;

	if (!node->name || strcasecmp(node->name, "Settings") != 0)
		return 1;

	prefs = xmlGetProp(node, "DefaultDir");
    if (prefs)
    {
        project->default_path = prefs;
    }
    else
    {
        g_string_printf(buffer, "%s/downloads/%s/", g_get_home_dir(),
                        project->name);
        project->default_path = g_strdup(buffer->str);
    }

	prefs = xmlGetProp(node, "MaxDownloads");
	if (prefs)
    {
		project->max_downloads = (unsigned int) atoi(prefs);
        xmlFree(prefs);
    }
	else
    {
		project->max_downloads = 1;
    }

	prefs = xmlGetProp(node, "AutoDownload");
	if (prefs)
    {
		project->auto_download = (unsigned int) atoi(prefs);
        xmlFree(prefs);
    }
	else
    {
		project->auto_download = 1;
    }

	xmlFreeDoc(doc);
	g_string_free(buffer, TRUE);

	return 0;
}

void
project_auto_set(Project *project, unsigned int setting, GError **error)
{
	project->auto_download = setting;
    g_signal_emit (G_OBJECT (project),
                   project_signals[PROJECT_AUTODOWNLOAD],
                   0);
}

void
project_maxdownloads_set (Project *project, guint maxdownloads, GError **error)
{
    g_print("maxdownloads set to %i\n", maxdownloads);
	project->max_downloads = maxdownloads;
    g_signal_emit (G_OBJECT (project),
                   project_signals[PROJECT_MAXDOWNLOADS],
                   0);
}

void
project_name_set (Project *project, const gchar *name)
{
    if (project->name)
        g_free (project->name);

    project->name = g_strdup (name);
}

const gchar*
project_name_get (Project *project)
{
    return project->name;
}

void
project_path_set (Project *project, const gchar *path)
{
    if (project->path)
        g_free (project->path);

    project->path = g_strdup (path);
}

const gchar*
project_path_get (Project *project)
{
    return project->path;
}

void
project_default_path_set (Project *project, const gchar *default_path)
{
    if (project->default_path)
        g_free (project->default_path);

    project->default_path = g_strdup (default_path);
}

const gchar*
project_default_path_get (Project *project)
{
    return project->default_path;
}

void
project_speedlimit_set (Project *project, guint speedlimit)
{
    project->speedlimit = speedlimit;
}

GHashTable *
project_hash (Project *project)
{
    return project->hash;
}

List *project_batch(Project *project)
{
	return project->batch;
}

List *project_batch_active(Project *project)
{
	return project->batch_active;
}

unsigned int project_size(Project *project)
{
	return list_size(project->batch);
}

unsigned int project_queued_downloads(Project *project)
{
	return project->queued_downloads;
}

void
project_lock_batch (Project *project)
{
    g_static_mutex_lock (&project->batch_mutex);
}

void
project_unlock_batch (Project *project)
{
    g_static_mutex_unlock (&project->batch_mutex);
}

void
project_lock_active (Project *project)
{
    g_static_mutex_lock (&project->batch_active_mutex);
}

void
project_unlock_active (Project *project)
{
    g_static_mutex_unlock (&project->batch_active_mutex);
}

int project_auto(Project *project)
{
	return project->auto_download;
}

guint project_maxdownloads_get (Project *project)
{
	return project->max_downloads;
}
